@@ -17,11 +17,11 @@ import traceback
 
 from . import __version__
 from .helpers import Error, location_validator, format_time, format_file_size, \
-    format_file_mode, ExcludePattern, IncludePattern, exclude_path, adjust_patterns, to_localtime, timestamp, \
+    format_file_mode, parse_pattern, PathPrefixPattern, to_localtime, timestamp, \
     get_cache_dir, get_keys_dir, prune_within, prune_split, unhexlify, \
     Manifest, remove_surrogates, update_excludes, format_archive, check_extension_modules, Statistics, \
     dir_is_tagged, bigint_to_int, ChunkerParams, CompressionSpec, is_slow_msgpack, yes, sysinfo, \
-    EXIT_SUCCESS, EXIT_WARNING, EXIT_ERROR
+    EXIT_SUCCESS, EXIT_WARNING, EXIT_ERROR, log_multi, PatternMatcher
 from .logger import create_logger, setup_logging
 logger = create_logger()
 from .compress import Compressor, COMPR_BUFFER
@@ -30,13 +30,15 @@ from .repository import Repository
 from .cache import Cache
 from .key import key_creator
 from .archive import Archive, ArchiveChecker, CHUNKER_PARAMS
-from .remote import RepositoryServer, RemoteRepository
+from .remote import RepositoryServer, RemoteRepository, cache_if_remote
 
 has_lchflags = hasattr(os, 'lchflags')
 
 # default umask, overriden by --umask, defaults to read/write only for owner
 UMASK_DEFAULT = 0o077
 
+DASHES = '-' * 78
+
 
 class ToggleAction(argparse.Action):
     """argparse action to handle "toggle" flags easily
@@ -78,7 +80,7 @@ class Archiver:
         logger.warning(msg)
 
     def print_file_status(self, status, path):
-        if self.output_filter is None or status in self.output_filter:
+        if self.output_list and (self.output_filter is None or status in self.output_filter):
             logger.info("%1s %s", status, remove_surrogates(path))
 
     def do_serve(self, args):
@@ -95,7 +97,8 @@ class Archiver:
         manifest.key = key
         manifest.write()
         repository.commit()
-        Cache(repository, key, manifest, warn_if_unencrypted=False)
+        with Cache(repository, key, manifest, warn_if_unencrypted=False):
+            pass
         return self.exit_code
 
     def do_check(self, args):
@@ -126,7 +129,65 @@ class Archiver:
 
     def do_create(self, args):
         """Create new archive"""
+        matcher = PatternMatcher(fallback=True)
+        if args.excludes:
+            matcher.add(args.excludes, False)
+
+        def create_inner(archive, cache):
+            # Add cache dir to inode_skip list
+            skip_inodes = set()
+            try:
+                st = os.stat(get_cache_dir())
+                skip_inodes.add((st.st_ino, st.st_dev))
+            except IOError:
+                pass
+            # Add local repository dir to inode_skip list
+            if not args.location.host:
+                try:
+                    st = os.stat(args.location.path)
+                    skip_inodes.add((st.st_ino, st.st_dev))
+                except IOError:
+                    pass
+            for path in args.paths:
+                if path == '-':  # stdin
+                    path = 'stdin'
+                    if not dry_run:
+                        try:
+                            status = archive.process_stdin(path, cache)
+                        except IOError as e:
+                            status = 'E'
+                            self.print_warning('%s: %s', path, e)
+                    else:
+                        status = '-'
+                    self.print_file_status(status, path)
+                    continue
+                path = os.path.normpath(path)
+                if args.one_file_system:
+                    try:
+                        restrict_dev = os.lstat(path).st_dev
+                    except OSError as e:
+                        self.print_warning('%s: %s', path, e)
+                        continue
+                else:
+                    restrict_dev = None
+                self._process(archive, cache, matcher, args.exclude_caches, args.exclude_if_present,
+                              args.keep_tag_files, skip_inodes, path, restrict_dev,
+                              read_special=args.read_special, dry_run=dry_run)
+            if not dry_run:
+                archive.save(timestamp=args.timestamp)
+                if args.progress:
+                    archive.stats.show_progress(final=True)
+                if args.stats:
+                    archive.end = datetime.now()
+                    log_multi(DASHES,
+                              str(archive),
+                              DASHES,
+                              str(archive.stats),
+                              str(cache),
+                              DASHES)
+
         self.output_filter = args.output_filter
+        self.output_list = args.output_list
         dry_run = args.dry_run
         t0 = datetime.now()
         if not dry_run:
@@ -135,71 +196,22 @@ class Archiver:
             compr_args = dict(buffer=COMPR_BUFFER)
             compr_args.update(args.compression)
             key.compressor = Compressor(**compr_args)
-            cache = Cache(repository, key, manifest, do_files=args.cache_files, lock_wait=self.lock_wait)
-            archive = Archive(repository, key, manifest, args.location.archive, cache=cache,
-                              create=True, checkpoint_interval=args.checkpoint_interval,
-                              numeric_owner=args.numeric_owner, progress=args.progress,
-                              chunker_params=args.chunker_params, start=t0)
+            with Cache(repository, key, manifest, do_files=args.cache_files, lock_wait=self.lock_wait) as cache:
+                archive = Archive(repository, key, manifest, args.location.archive, cache=cache,
+                                  create=True, checkpoint_interval=args.checkpoint_interval,
+                                  numeric_owner=args.numeric_owner, progress=args.progress,
+                                  chunker_params=args.chunker_params, start=t0)
+                create_inner(archive, cache)
         else:
-            archive = cache = None
-        # Add cache dir to inode_skip list
-        skip_inodes = set()
-        try:
-            st = os.stat(get_cache_dir())
-            skip_inodes.add((st.st_ino, st.st_dev))
-        except IOError:
-            pass
-        # Add local repository dir to inode_skip list
-        if not args.location.host:
-            try:
-                st = os.stat(args.location.path)
-                skip_inodes.add((st.st_ino, st.st_dev))
-            except IOError:
-                pass
-        for path in args.paths:
-            if path == '-':  # stdin
-                path = 'stdin'
-                if not dry_run:
-                    try:
-                        status = archive.process_stdin(path, cache)
-                    except IOError as e:
-                        status = 'E'
-                        self.print_warning('%s: %s', path, e)
-                else:
-                    status = '-'
-                self.print_file_status(status, path)
-                continue
-            path = os.path.normpath(path)
-            if args.one_file_system:
-                try:
-                    restrict_dev = os.lstat(path).st_dev
-                except OSError as e:
-                    self.print_warning('%s: %s', path, e)
-                    continue
-            else:
-                restrict_dev = None
-            self._process(archive, cache, args.excludes, args.exclude_caches, args.exclude_if_present,
-                          args.keep_tag_files, skip_inodes, path, restrict_dev,
-                          read_special=args.read_special, dry_run=dry_run)
-        if not dry_run:
-            archive.save(timestamp=args.timestamp)
-            if args.progress:
-                archive.stats.show_progress(final=True)
-            if args.stats:
-                archive.end = datetime.now()
-                print('-' * 78)
-                print(str(archive))
-                print()
-                print(str(archive.stats))
-                print(str(cache))
-                print('-' * 78)
+            create_inner(None, None)
         return self.exit_code
 
-    def _process(self, archive, cache, excludes, exclude_caches, exclude_if_present,
+    def _process(self, archive, cache, matcher, exclude_caches, exclude_if_present,
                  keep_tag_files, skip_inodes, path, restrict_dev,
                  read_special=False, dry_run=False):
-        if exclude_path(path, excludes):
+        if not matcher.match(path):
             return
+
         try:
             st = os.lstat(path)
         except OSError as e:
@@ -225,10 +237,10 @@ class Archiver:
         elif stat.S_ISDIR(st.st_mode):
             tag_paths = dir_is_tagged(path, exclude_caches, exclude_if_present)
             if tag_paths:
-                if keep_tag_files:
+                if keep_tag_files and not dry_run:
                     archive.process_dir(path, st)
                     for tag_path in tag_paths:
-                        self._process(archive, cache, excludes, exclude_caches, exclude_if_present,
+                        self._process(archive, cache, matcher, exclude_caches, exclude_if_present,
                                       keep_tag_files, skip_inodes, tag_path, restrict_dev,
                                       read_special=read_special, dry_run=dry_run)
                 return
@@ -242,7 +254,7 @@ class Archiver:
             else:
                 for filename in sorted(entries):
                     entry_path = os.path.normpath(os.path.join(path, filename))
-                    self._process(archive, cache, excludes, exclude_caches, exclude_if_present,
+                    self._process(archive, cache, matcher, exclude_caches, exclude_if_present,
                                   keep_tag_files, skip_inodes, entry_path, restrict_dev,
                                   read_special=read_special, dry_run=dry_run)
         elif stat.S_ISLNK(st.st_mode):
@@ -279,13 +291,25 @@ class Archiver:
         manifest, key = Manifest.load(repository)
         archive = Archive(repository, key, manifest, args.location.archive,
                           numeric_owner=args.numeric_owner)
-        patterns = adjust_patterns(args.paths, args.excludes)
+
+        matcher = PatternMatcher()
+        if args.excludes:
+            matcher.add(args.excludes, False)
+
+        include_patterns = []
+
+        if args.paths:
+            include_patterns.extend(parse_pattern(i, PathPrefixPattern) for i in args.paths)
+            matcher.add(include_patterns, True)
+
+        matcher.fallback = not include_patterns
+
         dry_run = args.dry_run
         stdout = args.stdout
         sparse = args.sparse
         strip_components = args.strip_components
         dirs = []
-        for item in archive.iter_items(lambda item: not exclude_path(item[b'path'], patterns), preload=True):
+        for item in archive.iter_items(lambda item: matcher.match(item[b'path']), preload=True):
             orig_path = item[b'path']
             if strip_components:
                 item[b'path'] = os.sep.join(orig_path.split(os.sep)[strip_components:])
@@ -310,8 +334,8 @@ class Archiver:
         if not args.dry_run:
             while dirs:
                 archive.extract_item(dirs.pop(-1))
-        for pattern in (patterns or []):
-            if isinstance(pattern, IncludePattern) and  pattern.match_count == 0:
+        for pattern in include_patterns:
+            if pattern.match_count == 0:
                 self.print_warning("Include pattern '%s' never matched.", pattern)
         return self.exit_code
 
@@ -319,45 +343,48 @@ class Archiver:
         """Rename an existing archive"""
         repository = self.open_repository(args, exclusive=True)
         manifest, key = Manifest.load(repository)
-        cache = Cache(repository, key, manifest, lock_wait=self.lock_wait)
-        archive = Archive(repository, key, manifest, args.location.archive, cache=cache)
-        archive.rename(args.name)
-        manifest.write()
-        repository.commit()
-        cache.commit()
+        with Cache(repository, key, manifest, lock_wait=self.lock_wait) as cache:
+            archive = Archive(repository, key, manifest, args.location.archive, cache=cache)
+            archive.rename(args.name)
+            manifest.write()
+            repository.commit()
+            cache.commit()
         return self.exit_code
 
     def do_delete(self, args):
         """Delete an existing repository or archive"""
         repository = self.open_repository(args, exclusive=True)
         manifest, key = Manifest.load(repository)
-        cache = Cache(repository, key, manifest, do_files=args.cache_files, lock_wait=self.lock_wait)
-        if args.location.archive:
-            archive = Archive(repository, key, manifest, args.location.archive, cache=cache)
-            stats = Statistics()
-            archive.delete(stats)
-            manifest.write()
-            repository.commit(save_space=args.save_space)
-            cache.commit()
-            if args.stats:
-                logger.info(stats.summary.format(label='Deleted data:', stats=stats))
-                logger.info(str(cache))
-        else:
-            if not args.cache_only:
-                msg = []
-                msg.append("You requested to completely DELETE the repository *including* all archives it contains:")
-                for archive_info in manifest.list_archive_infos(sort_by='ts'):
-                    msg.append(format_archive(archive_info))
-                msg.append("Type 'YES' if you understand this and want to continue: ")
-                msg = '\n'.join(msg)
-                if not yes(msg, false_msg="Aborting.", default_notty=False,
-                           env_var_override='BORG_CHECK_I_KNOW_WHAT_I_AM_DOING', truish=('YES', )):
-                    self.exit_code = EXIT_ERROR
-                    return self.exit_code
-                repository.destroy()
-                logger.info("Repository deleted.")
-            cache.destroy()
-            logger.info("Cache deleted.")
+        with Cache(repository, key, manifest, do_files=args.cache_files, lock_wait=self.lock_wait) as cache:
+            if args.location.archive:
+                archive = Archive(repository, key, manifest, args.location.archive, cache=cache)
+                stats = Statistics()
+                archive.delete(stats, progress=args.progress)
+                manifest.write()
+                repository.commit(save_space=args.save_space)
+                cache.commit()
+                logger.info("Archive deleted.")
+                if args.stats:
+                    log_multi(DASHES,
+                              stats.summary.format(label='Deleted data:', stats=stats),
+                              str(cache),
+                              DASHES)
+            else:
+                if not args.cache_only:
+                    msg = []
+                    msg.append("You requested to completely DELETE the repository *including* all archives it contains:")
+                    for archive_info in manifest.list_archive_infos(sort_by='ts'):
+                        msg.append(format_archive(archive_info))
+                    msg.append("Type 'YES' if you understand this and want to continue: ")
+                    msg = '\n'.join(msg)
+                    if not yes(msg, false_msg="Aborting.", default_notty=False,
+                               env_var_override='BORG_DELETE_I_KNOW_WHAT_I_AM_DOING', truish=('YES', )):
+                        self.exit_code = EXIT_ERROR
+                        return self.exit_code
+                    repository.destroy()
+                    logger.info("Repository deleted.")
+                cache.destroy()
+                logger.info("Cache deleted.")
         return self.exit_code
 
     def do_mount(self, args):
@@ -374,18 +401,19 @@ class Archiver:
 
         repository = self.open_repository(args)
         try:
-            manifest, key = Manifest.load(repository)
-            if args.location.archive:
-                archive = Archive(repository, key, manifest, args.location.archive)
-            else:
-                archive = None
-            operations = FuseOperations(key, repository, manifest, archive)
-            logger.info("Mounting filesystem")
-            try:
-                operations.mount(args.mountpoint, args.options, args.foreground)
-            except RuntimeError:
-                # Relevant error message already printed to stderr by fuse
-                self.exit_code = EXIT_ERROR
+            with cache_if_remote(repository) as cached_repo:
+                manifest, key = Manifest.load(repository)
+                if args.location.archive:
+                    archive = Archive(repository, key, manifest, args.location.archive)
+                else:
+                    archive = None
+                operations = FuseOperations(key, repository, manifest, archive, cached_repo)
+                logger.info("Mounting filesystem")
+                try:
+                    operations.mount(args.mountpoint, args.options, args.foreground)
+                except RuntimeError:
+                    # Relevant error message already printed to stderr by fuse
+                    self.exit_code = EXIT_ERROR
         finally:
             repository.close()
         return self.exit_code
@@ -438,29 +466,28 @@ class Archiver:
         """Show archive details such as disk space used"""
         repository = self.open_repository(args)
         manifest, key = Manifest.load(repository)
-        cache = Cache(repository, key, manifest, do_files=args.cache_files, lock_wait=self.lock_wait)
-        archive = Archive(repository, key, manifest, args.location.archive, cache=cache)
-        stats = archive.calc_stats(cache)
-        print('Name:', archive.name)
-        print('Fingerprint: %s' % hexlify(archive.id).decode('ascii'))
-        print('Hostname:', archive.metadata[b'hostname'])
-        print('Username:', archive.metadata[b'username'])
-        print('Time: %s' % format_time(to_localtime(archive.ts)))
-        print('Command line:', remove_surrogates(' '.join(archive.metadata[b'cmdline'])))
-        print('Number of files: %d' % stats.nfiles)
-        print()
-        print(str(stats))
-        print(str(cache))
+        with Cache(repository, key, manifest, do_files=args.cache_files, lock_wait=self.lock_wait) as cache:
+            archive = Archive(repository, key, manifest, args.location.archive, cache=cache)
+            stats = archive.calc_stats(cache)
+            print('Name:', archive.name)
+            print('Fingerprint: %s' % hexlify(archive.id).decode('ascii'))
+            print('Hostname:', archive.metadata[b'hostname'])
+            print('Username:', archive.metadata[b'username'])
+            print('Time: %s' % format_time(to_localtime(archive.ts)))
+            print('Command line:', remove_surrogates(' '.join(archive.metadata[b'cmdline'])))
+            print('Number of files: %d' % stats.nfiles)
+            print()
+            print(str(stats))
+            print(str(cache))
         return self.exit_code
 
     def do_prune(self, args):
         """Prune repository archives according to specified rules"""
         repository = self.open_repository(args, exclusive=True)
         manifest, key = Manifest.load(repository)
-        cache = Cache(repository, key, manifest, do_files=args.cache_files, lock_wait=self.lock_wait)
         archives = manifest.list_archive_infos(sort_by='ts', reverse=True)  # just a ArchiveInfo list
         if args.hourly + args.daily + args.weekly + args.monthly + args.yearly == 0 and args.within is None:
-            self.print_error('At least one of the "within", "keep-hourly", "keep-daily", "keep-weekly", '
+            self.print_error('At least one of the "keep-within", "keep-hourly", "keep-daily", "keep-weekly", '
                              '"keep-monthly" or "keep-yearly" settings must be specified')
             return self.exit_code
         if args.prefix:
@@ -482,21 +509,24 @@ class Archiver:
         keep.sort(key=attrgetter('ts'), reverse=True)
         to_delete = [a for a in archives if a not in keep]
         stats = Statistics()
-        for archive in keep:
-            logger.info('Keeping archive: %s' % format_archive(archive))
-        for archive in to_delete:
-            if args.dry_run:
-                logger.info('Would prune:     %s' % format_archive(archive))
-            else:
-                logger.info('Pruning archive: %s' % format_archive(archive))
-                Archive(repository, key, manifest, archive.name, cache).delete(stats)
-        if to_delete and not args.dry_run:
-            manifest.write()
-            repository.commit(save_space=args.save_space)
-            cache.commit()
-        if args.stats:
-            logger.info(stats.summary.format(label='Deleted data:', stats=stats))
-            logger.info(str(cache))
+        with Cache(repository, key, manifest, do_files=args.cache_files, lock_wait=self.lock_wait) as cache:
+            for archive in keep:
+                logger.info('Keeping archive: %s' % format_archive(archive))
+            for archive in to_delete:
+                if args.dry_run:
+                    logger.info('Would prune:     %s' % format_archive(archive))
+                else:
+                    logger.info('Pruning archive: %s' % format_archive(archive))
+                    Archive(repository, key, manifest, archive.name, cache).delete(stats)
+            if to_delete and not args.dry_run:
+                manifest.write()
+                repository.commit(save_space=args.save_space)
+                cache.commit()
+            if args.stats:
+                log_multi(DASHES,
+                          stats.summary.format(label='Deleted data:', stats=stats),
+                          str(cache),
+                          DASHES)
         return self.exit_code
 
     def do_upgrade(self, args):
@@ -512,7 +542,7 @@ class Archiver:
         # XXX: should auto-detect if it is an attic repository here
         repo = AtticRepositoryUpgrader(args.location.path, create=False)
         try:
-            repo.upgrade(args.dry_run, inplace=args.inplace)
+            repo.upgrade(args.dry_run, inplace=args.inplace, progress=args.progress)
         except NotImplementedError as e:
             print("warning: %s" % e)
         return self.exit_code
@@ -597,18 +627,62 @@ class Archiver:
         return self.exit_code
 
     helptext = {}
-    helptext['patterns'] = '''
-        Exclude patterns use a variant of shell pattern syntax, with '*' matching any
-        number of characters, '?' matching any single character, '[...]' matching any
-        single character specified, including ranges, and '[!...]' matching any
-        character not specified.  For the purpose of these patterns, the path
-        separator ('\\' for Windows and '/' on other systems) is not treated
-        specially.  For a path to match a pattern, it must completely match from
-        start to end, or must match from the start to just before a path separator.
-        Except for the root path, paths will never end in the path separator when
-        matching is attempted.  Thus, if a given pattern ends in a path separator, a
-        '*' is appended before matching is attempted.  Patterns with wildcards should
-        be quoted to protect them from shell expansion.
+    helptext['patterns'] = textwrap.dedent('''
+        Exclusion patterns support four separate styles, fnmatch, shell, regular
+        expressions and path prefixes. If followed by a colon (':') the first two
+        characters of a pattern are used as a style selector. Explicit style
+        selection is necessary when a non-default style is desired or when the
+        desired pattern starts with two alphanumeric characters followed by a colon
+        (i.e. `aa:something/*`).
+
+        `Fnmatch <https://docs.python.org/3/library/fnmatch.html>`_, selector `fm:`
+
+            These patterns use a variant of shell pattern syntax, with '*' matching
+            any number of characters, '?' matching any single character, '[...]'
+            matching any single character specified, including ranges, and '[!...]'
+            matching any character not specified. For the purpose of these patterns,
+            the path separator ('\\' for Windows and '/' on other systems) is not
+            treated specially. Wrap meta-characters in brackets for a literal match
+            (i.e. `[?]` to match the literal character `?`). For a path to match
+            a pattern, it must completely match from start to end, or must match from
+            the start to just before a path separator. Except for the root path,
+            paths will never end in the path separator when matching is attempted.
+            Thus, if a given pattern ends in a path separator, a '*' is appended
+            before matching is attempted.
+
+        Shell-style patterns, selector `sh:`
+
+            Like fnmatch patterns these are similar to shell patterns. The difference
+            is that the pattern may include `**/` for matching zero or more directory
+            levels, `*` for matching zero or more arbitrary characters with the
+            exception of any path separator.
+
+        Regular expressions, selector `re:`
+
+            Regular expressions similar to those found in Perl are supported. Unlike
+            shell patterns regular expressions are not required to match the complete
+            path and any substring match is sufficient. It is strongly recommended to
+            anchor patterns to the start ('^'), to the end ('$') or both. Path
+            separators ('\\' for Windows and '/' on other systems) in paths are
+            always normalized to a forward slash ('/') before applying a pattern. The
+            regular expression syntax is described in the `Python documentation for
+            the re module <https://docs.python.org/3/library/re.html>`_.
+
+        Prefix path, selector `pp:`
+
+            This pattern style is useful to match whole sub-directories. The pattern
+            `pp:/data/bar` matches `/data/bar` and everything therein.
+
+        Exclusions can be passed via the command line option `--exclude`. When used
+        from within a shell the patterns should be quoted to protect them from
+        expansion.
+
+        The `--exclude-from` option permits loading exclusion patterns from a text
+        file with one pattern per line. Lines empty or starting with the number sign
+        ('#') after removing whitespace on both ends are ignored. The optional style
+        selector prefix is also supported for patterns loaded from a file. Due to
+        whitespace removal paths with whitespace at the beginning or end can only be
+        excluded using regular expressions.
 
         Examples:
 
@@ -624,7 +698,22 @@ class Archiver:
 
         # The file '/home/user/cache/important' is *not* backed up:
         $ borg create -e /home/user/cache/ backup / /home/user/cache/important
-        '''
+
+        # The contents of directories in '/home' are not backed up when their name
+        # ends in '.tmp'
+        $ borg create --exclude 're:^/home/[^/]+\.tmp/' backup /
+
+        # Load exclusions from file
+        $ cat >exclude.txt <<EOF
+        # Comment line
+        /home/*/junk
+        *.tmp
+        fm:aa:something/*
+        re:^/home/[^/]\.tmp/
+        sh:/home/*/.thumbnails
+        EOF
+        $ borg create --exclude-from exclude.txt backup /
+        ''')
 
     def do_help(self, parser, commands, args):
         if not args.topic:
@@ -809,10 +898,13 @@ class Archiver:
                                help="""show progress display while creating the archive, showing Original,
                                Compressed and Deduplicated sizes, followed by the Number of files seen
                                and the path being processed, default: %(default)s""")
+        subparser.add_argument('--list', dest='output_list',
+                               action='store_true', default=False,
+                               help='output verbose list of items (files, dirs, ...)')
         subparser.add_argument('--filter', dest='output_filter', metavar='STATUSCHARS',
                                help='only display items with the given status characters')
         subparser.add_argument('-e', '--exclude', dest='excludes',
-                               type=ExcludePattern, action='append',
+                               type=parse_pattern, action='append',
                                metavar="PATTERN", help='exclude paths matching PATTERN')
         subparser.add_argument('--exclude-from', dest='exclude_files',
                                type=argparse.FileType('r'), action='append',
@@ -882,7 +974,7 @@ class Archiver:
                                default=False, action='store_true',
                                help='do not actually change any files')
         subparser.add_argument('-e', '--exclude', dest='excludes',
-                               type=ExcludePattern, action='append',
+                               type=parse_pattern, action='append',
                                metavar="PATTERN", help='exclude paths matching PATTERN')
         subparser.add_argument('--exclude-from', dest='exclude_files',
                                type=argparse.FileType('r'), action='append',
@@ -903,7 +995,7 @@ class Archiver:
                                type=location_validator(archive=True),
                                help='archive to extract')
         subparser.add_argument('paths', metavar='PATH', nargs='*', type=str,
-                               help='paths to extract')
+                               help='paths to extract; patterns are supported')
 
         rename_epilog = textwrap.dedent("""
         This command renames an archive in the repository.
@@ -929,6 +1021,9 @@ class Archiver:
                                           epilog=delete_epilog,
                                           formatter_class=argparse.RawDescriptionHelpFormatter)
         subparser.set_defaults(func=self.do_delete)
+        subparser.add_argument('-p', '--progress', dest='progress',
+                               action='store_true', default=False,
+                               help="""show progress display while deleting a single archive""")
         subparser.add_argument('-s', '--stats', dest='stats',
                                action='store_true', default=False,
                                help='print statistics for the deleted archive')
@@ -1063,46 +1158,49 @@ class Archiver:
                                help='repository to prune')
 
         upgrade_epilog = textwrap.dedent("""
-        upgrade an existing Borg repository. this currently
-        only support converting an Attic repository, but may
+        Upgrade an existing Borg repository. This currently
+        only supports converting an Attic repository, but may
         eventually be extended to cover major Borg upgrades as well.
 
-        it will change the magic strings in the repository's segments
-        to match the new Borg magic strings. the keyfiles found in
+        It will change the magic strings in the repository's segments
+        to match the new Borg magic strings. The keyfiles found in
         $ATTIC_KEYS_DIR or ~/.attic/keys/ will also be converted and
         copied to $BORG_KEYS_DIR or ~/.borg/keys.
 
-        the cache files are converted, from $ATTIC_CACHE_DIR or
+        The cache files are converted, from $ATTIC_CACHE_DIR or
         ~/.cache/attic to $BORG_CACHE_DIR or ~/.cache/borg, but the
         cache layout between Borg and Attic changed, so it is possible
         the first backup after the conversion takes longer than expected
         due to the cache resync.
 
-        upgrade should be able to resume if interrupted, although it
-        will still iterate over all segments. if you want to start
+        Upgrade should be able to resume if interrupted, although it
+        will still iterate over all segments. If you want to start
         from scratch, use `borg delete` over the copied repository to
         make sure the cache files are also removed:
 
             borg delete borg
 
-        unless ``--inplace`` is specified, the upgrade process first
+        Unless ``--inplace`` is specified, the upgrade process first
         creates a backup copy of the repository, in
-        REPOSITORY.upgrade-DATETIME, using hardlinks. this takes
+        REPOSITORY.upgrade-DATETIME, using hardlinks. This takes
         longer than in place upgrades, but is much safer and gives
-        progress information (as opposed to ``cp -al``). once you are
+        progress information (as opposed to ``cp -al``). Once you are
         satisfied with the conversion, you can safely destroy the
         backup copy.
 
-        WARNING: running the upgrade in place will make the current
+        WARNING: Running the upgrade in place will make the current
         copy unusable with older version, with no way of going back
-        to previous versions. this can PERMANENTLY DAMAGE YOUR
+        to previous versions. This can PERMANENTLY DAMAGE YOUR
         REPOSITORY!  Attic CAN NOT READ BORG REPOSITORIES, as the
-        magic strings have changed. you have been warned.""")
+        magic strings have changed. You have been warned.""")
         subparser = subparsers.add_parser('upgrade', parents=[common_parser],
                                           description=self.do_upgrade.__doc__,
                                           epilog=upgrade_epilog,
                                           formatter_class=argparse.RawDescriptionHelpFormatter)
         subparser.set_defaults(func=self.do_upgrade)
+        subparser.add_argument('-p', '--progress', dest='progress',
+                               action='store_true', default=False,
+                               help="""show progress display while upgrading the repository""")
         subparser.add_argument('-n', '--dry-run', dest='dry_run',
                                default=False, action='store_true',
                                help='do not change repository')
